// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#ifndef __ABG_VERSION_H__
#define __ABG_VERSION_H__
#define ABIGAIL_VERSION_MAJOR "2"
#define ABIGAIL_VERSION_MINOR "10"
#define ABIGAIL_VERSION_REVISION "0"
#define ABIGAIL_VERSION_SUFFIX ""
#define ABIGAIL_ABIXML_VERSION_MAJOR "2"
#define ABIGAIL_ABIXML_VERSION_MINOR "4"
#endif
